"""
This module will parse the JSON file following the BNF definition:

    <json> ::= <container>

    <primitive> ::= <number> | <string> | <boolean>
    ; Where:
    ; <number> is a valid real number expressed in one of a number of given formats
    ; <string> is a string of valid characters enclosed in quotes
    ; <boolean> is one of the literal strings 'true', 'false', or 'null' (unquoted)

    <container> ::= <object> | <array>
    <array> ::= '[' [ <json> *(', ' <json>) ] ']' ; A sequence of JSON values separated by commas
    <object> ::= '{' [ <member> *(', ' <member>) ] '}' ; A sequence of 'members'
    <member> ::= <string> ': ' <json> ; A pair consisting of a name, and a JSON value

If something is wrong (a missing parentheses or quotes for example) it will use a few simple heuristics to fix the JSON string:
- Add the missing parentheses if the parser believes that the array or object should be closed
- Quote strings or add missing single quotes
- Adjust whitespaces and remove line breaks

All supported use cases are in the unit tests
"""

import argparse
import json
import sys
from pathlib import Path
from typing import Any, Literal, TextIO, overload

from .json_parser import JSONParser
from .schema_repair import SchemaRepairer, load_schema_model, schema_from_input
from .utils.constants import JSONReturnType


@overload
def repair_json(
    json_str: str = "",
    return_objects: Literal[False] = False,
    skip_json_loads: bool = False,
    logging: bool = False,
    json_fd: TextIO | None = None,
    chunk_length: int = 0,
    stream_stable: bool = False,
    strict: bool = False,
    schema: Any | None = None,
    **json_dumps_args: Any,
) -> str: ...


@overload
def repair_json(
    json_str: str = "",
    return_objects: Literal[True] = True,
    skip_json_loads: bool = False,
    logging: bool = False,
    json_fd: TextIO | None = None,
    chunk_length: int = 0,
    stream_stable: bool = False,
    strict: bool = False,
    schema: Any | None = None,
    **json_dumps_args: Any,
) -> JSONReturnType | tuple[JSONReturnType, list[dict[str, str]]]: ...


def repair_json(
    json_str: str = "",
    return_objects: bool = False,
    skip_json_loads: bool = False,
    logging: bool = False,
    json_fd: TextIO | None = None,
    chunk_length: int = 0,
    stream_stable: bool = False,
    strict: bool = False,
    schema: Any | None = None,
    **json_dumps_args: Any,
) -> JSONReturnType | tuple[JSONReturnType, list[dict[str, str]]]:
    """
    Given a json formatted string, it will try to decode it and, if it fails, it will try to fix it.

    Args:
        json_str (str, optional): The JSON string to repair. Defaults to an empty string.
        return_objects (bool, optional): If True, return the decoded data structure. Defaults to False.
        skip_json_loads (bool, optional): If True, skip calling the built-in json.loads() function to verify that the json is valid before attempting to repair. Defaults to False.
        logging (bool, optional): If True, return a tuple with the repaired json and a log of all repair actions. Defaults to False. When no repairs were required, the repair log will be an empty list.
        json_fd (Optional[TextIO], optional): File descriptor for JSON input. Do not use! Use `from_file` or `load` instead. Defaults to None.
        ensure_ascii (bool, optional): Set to False to avoid converting non-latin characters to ascii (for example when using chinese characters). Defaults to True. Ignored if `skip_json_loads` is True.
        chunk_length (int, optional): Size in bytes of the file chunks to read at once. Ignored if `json_fd` is None. Do not use! Use `from_file` or `load` instead. Defaults to 1MB.
        stream_stable (bool, optional): When the json to be repaired is the accumulation of streaming json at a certain moment.If this parameter to True will keep the repair results stable.
        strict (bool, optional): If True, surface structural problems (duplicate keys, missing separators, empty keys/values, etc.) as ValueError instead of repairing them.
        schema (Any, optional): JSON Schema dict, boolean schema, or pydantic v2 model used to guide repairs and validation for both valid and invalid JSON inputs.
    Returns:
        Union[JSONReturnType, Tuple[JSONReturnType, List[Dict[str, str]]]]: The repaired JSON or a tuple with the repaired JSON and repair log when logging is True.
    """
    # Schema-guided repairs and strict mode are mutually exclusive to avoid conflicting behavior.
    if schema is not None and strict:
        raise ValueError("schema and strict cannot be used together.")

    parser = JSONParser(json_str, json_fd, logging, chunk_length, stream_stable, strict)
    schema_obj = schema_from_input(schema) if schema is not None else None
    repairer = SchemaRepairer(schema_obj, parser.logger if logging else None) if schema_obj is not None else None

    # Fast path for valid JSON: schema-aware mode still applies repair+validation.
    if not skip_json_loads:
        try:
            loaded_json: JSONReturnType = json.load(json_fd) if json_fd else json.loads(json_str)
        except json.JSONDecodeError:
            pass
        else:
            if repairer is not None:
                assert schema_obj is not None
                repaired_json = repairer.repair_value(loaded_json, schema_obj, "$")
                repairer.validate(repaired_json, schema_obj)
                if logging:
                    return repaired_json, parser.logger
                if return_objects:
                    return repaired_json
                if repaired_json == "":
                    return ""
                return json.dumps(repaired_json, **json_dumps_args)
            if logging:
                return loaded_json, []
            if return_objects:
                return loaded_json
            if loaded_json == "":
                return ""
            return json.dumps(loaded_json, **json_dumps_args)

    parsed_json: JSONReturnType
    if repairer is None:
        parsed_json = parser.parse()
    else:
        assert schema_obj is not None
        parsed_json = parser.parse_with_schema(repairer, schema_obj)
        # Post-parse validation ensures we reject values that cannot satisfy the schema.
        repairer.validate(parsed_json, schema_obj)
    # It's useful to return the actual object instead of the json string,
    # it allows this lib to be a replacement of the json library
    if logging:
        return parsed_json, parser.logger
    if return_objects:
        return parsed_json
    # Avoid returning only a pair of quotes if it's an empty string
    if parsed_json == "":
        return ""
    return json.dumps(parsed_json, **json_dumps_args)


def loads(
    json_str: str,
    skip_json_loads: bool = False,
    logging: bool = False,
    stream_stable: bool = False,
    strict: bool = False,
    schema: Any | None = None,
) -> JSONReturnType | tuple[JSONReturnType, list[dict[str, str]]] | str:
    """
    This function works like `json.loads()` except that it will fix your JSON in the process.
    It is a wrapper around the `repair_json()` function with `return_objects=True`.

    Args:
        json_str (str): The JSON string to load and repair.
        skip_json_loads (bool, optional): If True, skip calling the built-in json.loads() function to verify that the json is valid before attempting to repair. Defaults to False.
        logging (bool, optional): If True, return a tuple with the repaired json and a log of all repair actions. Defaults to False.
        strict (bool, optional): If True, surface structural problems (duplicate keys, missing separators, empty keys/values, etc.) as ValueError instead of repairing them.
        schema (Any, optional): JSON Schema dict, boolean schema, or pydantic v2 model used to guide repairs and validation for both valid and invalid JSON inputs.

    Returns:
        Union[JSONReturnType, Tuple[JSONReturnType, List[Dict[str, str]]], str]: The repaired JSON object or a tuple with the repaired JSON object and repair log.
    """
    return repair_json(
        json_str=json_str,
        return_objects=True,
        skip_json_loads=skip_json_loads,
        logging=logging,
        stream_stable=stream_stable,
        strict=strict,
        schema=schema,
    )


def load(
    fd: TextIO,
    skip_json_loads: bool = False,
    logging: bool = False,
    chunk_length: int = 0,
    strict: bool = False,
    schema: Any | None = None,
) -> JSONReturnType | tuple[JSONReturnType, list[dict[str, str]]]:
    """
    This function works like `json.load()` except that it will fix your JSON in the process.
    It is a wrapper around the `repair_json()` function with `json_fd=fd` and `return_objects=True`.

    Args:
        fd (TextIO): File descriptor for JSON input.
        skip_json_loads (bool, optional): If True, skip calling the built-in json.loads() function to verify that the json is valid before attempting to repair. Defaults to False.
        logging (bool, optional): If True, return a tuple with the repaired json and a log of all repair actions. Defaults to False.
        chunk_length (int, optional): Size in bytes of the file chunks to read at once. Defaults to 1MB.
        strict (bool, optional): If True, surface structural problems (duplicate keys, missing separators, empty keys/values, etc.) as ValueError instead of repairing them.
        schema (Any, optional): JSON Schema dict, boolean schema, or pydantic v2 model used to guide repairs and validation for both valid and invalid JSON inputs.

    Returns:
        Union[JSONReturnType, Tuple[JSONReturnType, List[Dict[str, str]]]]: The repaired JSON object or a tuple with the repaired JSON object and repair log.
    """
    return repair_json(
        json_fd=fd,
        chunk_length=chunk_length,
        return_objects=True,
        skip_json_loads=skip_json_loads,
        logging=logging,
        strict=strict,
        schema=schema,
    )


def from_file(
    filename: str | Path,
    skip_json_loads: bool = False,
    logging: bool = False,
    chunk_length: int = 0,
    strict: bool = False,
    schema: Any | None = None,
) -> JSONReturnType | tuple[JSONReturnType, list[dict[str, str]]]:
    """
    This function is a wrapper around `load()` so you can pass the filename as string

    Args:
        filename (str | Path): The name of the file containing JSON data to load and repair.
        skip_json_loads (bool, optional): If True, skip calling the built-in json.loads() function to verify that the json is valid before attempting to repair. Defaults to False.
        logging (bool, optional): If True, return a tuple with the repaired json and a log of all repair actions. Defaults to False.
        chunk_length (int, optional): Size in bytes of the file chunks to read at once. Defaults to 1MB.
        strict (bool, optional): If True, surface structural problems (duplicate keys, missing separators, empty keys/values, etc.) as ValueError instead of repairing them.
        schema (Any, optional): JSON Schema dict, boolean schema, or pydantic v2 model used to guide repairs and validation for both valid and invalid JSON inputs.

    Returns:
        Union[JSONReturnType, Tuple[JSONReturnType, List[Dict[str, str]]]]: The repaired JSON object or a tuple with the repaired JSON object and repair log.
    """
    with Path(filename).open() as fd:
        return load(
            fd=fd,
            skip_json_loads=skip_json_loads,
            logging=logging,
            chunk_length=chunk_length,
            strict=strict,
            schema=schema,
        )


def cli(inline_args: list[str] | None = None) -> int:
    """
    Command-line interface for repairing and parsing JSON files.

    Args:
        inline_args (Optional[List[str]]): List of command-line arguments for testing purposes. Defaults to None.
            - filename (str): The JSON file to repair. If omitted, the JSON is read from stdin.
            - -i, --inline (bool): Replace the file inline instead of returning the output to stdout.
            - -o, --output TARGET (str): If specified, the output will be written to TARGET filename instead of stdout.
            - --ensure_ascii (bool): Pass ensure_ascii=True to json.dumps(). Will pass False otherwise.
            - --indent INDENT (int): Number of spaces for indentation (Default 2).
            - --skip-json-loads (bool): Skip initial json.loads validation.
            - --schema SCHEMA (str): Path to a JSON Schema file that guides repairs.
            - --schema-model MODEL (str): Pydantic v2 model in 'module:ClassName' form that guides repairs.
            - --strict (bool): Raise on duplicate keys, missing separators, empty keys/values, and other unrecoverable structures instead of repairing them.

    Returns:
        int: Exit code of the CLI operation.

    Raises:
        Exception: Any exception that occurs during file processing.

    Example:
        >>> cli(['example.json', '--indent', '4'])
        >>> cat json.txt | json_repair
    """
    parser = argparse.ArgumentParser(description="Repair and parse JSON files.")
    # Make the filename argument optional; if omitted, we will read from stdin.
    parser.add_argument(
        "filename",
        nargs="?",
        help="The JSON file to repair (if omitted, reads from stdin)",
    )
    parser.add_argument(
        "-i",
        "--inline",
        action="store_true",
        help="Replace the file inline instead of returning the output to stdout",
    )
    parser.add_argument(
        "-o",
        "--output",
        metavar="TARGET",
        help="If specified, the output will be written to TARGET filename instead of stdout",
    )
    parser.add_argument(
        "--ensure_ascii",
        action="store_true",
        help="Pass ensure_ascii=True to json.dumps()",
    )
    parser.add_argument(
        "--indent",
        type=int,
        default=2,
        help="Number of spaces for indentation (Default 2)",
    )
    parser.add_argument(
        "--skip-json-loads",
        action="store_true",
        help="Skip initial json.loads validation",
    )
    parser.add_argument(
        "--schema",
        metavar="SCHEMA",
        help="Path to a JSON Schema file that guides repairs",
    )
    parser.add_argument(
        "--schema-model",
        metavar="MODEL",
        help="Pydantic v2 model in 'module:ClassName' form that guides repairs",
    )
    parser.add_argument(
        "--strict",
        action="store_true",
        help="Raise on duplicate keys, missing separators, empty keys/values, and other unrecoverable structures instead of repairing them",
    )

    args = parser.parse_args(inline_args)

    # Inline mode requires a filename, so error out if none was provided.
    if args.inline and not args.filename:  # pragma: no cover
        print("Error: Inline mode requires a filename", file=sys.stderr)
        sys.exit(1)

    if args.inline and args.output:  # pragma: no cover
        print("Error: You cannot pass both --inline and --output", file=sys.stderr)
        sys.exit(1)

    if args.schema and args.schema_model:
        print("Error: You cannot pass both --schema and --schema-model", file=sys.stderr)
        sys.exit(1)

    if args.strict and (args.schema or args.schema_model):
        print("Error: --strict cannot be used with --schema or --schema-model", file=sys.stderr)
        sys.exit(1)

    ensure_ascii = args.ensure_ascii

    try:
        schema = None
        if args.schema:
            with Path(args.schema).open() as fd:
                schema = json.load(fd)
        elif args.schema_model:
            schema = load_schema_model(args.schema_model)

        # Use from_file if a filename is provided; otherwise read from stdin.
        if args.filename:
            result = from_file(
                args.filename,
                skip_json_loads=args.skip_json_loads,
                strict=args.strict,
                schema=schema,
            )
        else:
            data = sys.stdin.read()
            result = loads(
                data,
                skip_json_loads=args.skip_json_loads,
                strict=args.strict,
                schema=schema,
            )
        if args.inline or args.output:
            with Path(args.output or args.filename).open(mode="w") as fd:
                json.dump(result, fd, indent=args.indent, ensure_ascii=ensure_ascii)
        else:
            print(json.dumps(result, indent=args.indent, ensure_ascii=ensure_ascii))
    except (OSError, TypeError, ValueError) as e:  # pragma: no cover
        print(f"Error: {str(e)}", file=sys.stderr)
        return 1

    return 0  # Success


if __name__ == "__main__":  # pragma: no cover
    sys.exit(cli())



import pytest

from src.json_repair import repair_json


def repair_with_schema(raw, schema, **kwargs):
    return repair_json(raw, schema=schema, skip_json_loads=True, return_objects=True, **kwargs)


def test_schema_guides_missing_value_type_defaults():
    pytest.importorskip("jsonschema")
    schema = {
        "type": "object",
        "properties": {
            "text": {"type": "string"},
            "count": {"type": "integer"},
            "ratio": {"type": "number"},
            "flag": {"type": "boolean"},
            "items": {"type": "array", "items": {"type": "string"}},
            "payload": {"type": "object"},
            "nothing": {"type": "null"},
        },
        "required": ["text", "count", "ratio", "flag", "items", "payload", "nothing"],
    }
    raw = '{ "text": , "count": , "ratio": , "flag": , "items": , "payload": , "nothing": }'
    assert repair_with_schema(raw, schema) == {
        "text": "",
        "count": 0,
        "ratio": 0,
        "flag": False,
        "items": [],
        "payload": {},
        "nothing": None,
    }


def test_schema_missing_required_property_raises():
    pytest.importorskip("jsonschema")
    schema = {
        "type": "object",
        "properties": {"required_value": {"type": "integer", "default": 1}},
        "required": ["required_value"],
    }
    with pytest.raises(ValueError, match="Missing required properties"):
        repair_with_schema("{}", schema)


def test_schema_optional_default_is_inserted():
    pytest.importorskip("jsonschema")
    schema = {
        "type": "object",
        "properties": {"note": {"type": "string", "default": "n/a"}},
    }
    assert repair_with_schema("{}", schema) == {"note": "n/a"}


def test_schema_and_strict_are_mutually_exclusive():
    with pytest.raises(ValueError, match="schema and strict"):
        repair_json("{}", schema={}, strict=True, return_objects=True)


def test_schema_applies_to_valid_json_without_skip_json_loads():
    pytest.importorskip("jsonschema")
    with pytest.raises(ValueError, match="Expected string"):
        repair_json("true", schema={"type": "string"}, return_objects=True)

    schema = {
        "type": "object",
        "properties": {"value": {"type": "integer"}},
        "required": ["value"],
    }
    assert repair_json('{"value": "1"}', schema=schema, return_objects=True) == {"value": 1}

    with pytest.raises(ValueError, match="does not match"):
        repair_json('"bbb"', schema={"type": "string", "pattern": "^a+$"}, return_objects=True)


def test_schema_applies_to_valid_json_fast_path_outputs_and_logging():
    pytest.importorskip("jsonschema")
    schema = {
        "type": "object",
        "properties": {"value": {"type": "integer"}},
        "required": ["value"],
    }
    assert repair_json('{"value": "1"}', schema=schema) == '{"value": 1}'

    repaired, logs = repair_json('{"value": "1"}', schema=schema, logging=True)
    assert repaired == {"value": 1}
    assert logs


def test_schema_applies_to_valid_empty_string():
    pytest.importorskip("jsonschema")
    assert repair_json('""', schema={"type": "string"}) == ""


def test_schema_pydantic_v2_defaults():
    pydantic = pytest.importorskip("pydantic")
    version = getattr(pydantic, "VERSION", getattr(pydantic, "__version__", "0"))
    if int(version.split(".")[0]) < 2:
        pytest.skip("pydantic v2 required")

    base_model = pydantic.BaseModel
    field = pydantic.Field

    class SchemaModel(base_model):
        evidence_types: list[str] = field(default_factory=list)

    raw = '{ "evidence_types": }'
    assert repair_with_schema(raw, SchemaModel) == {"evidence_types": []}



import io
import json
import os
import tempfile
from pathlib import Path
from unittest.mock import patch

import pytest

from src.json_repair.json_repair import cli


def test_cli(capsys):
    # Create a temporary file
    temp_fd, temp_path = tempfile.mkstemp(suffix=".json")
    _, tempout_path = tempfile.mkstemp(suffix=".json")
    temp_path = Path(temp_path)
    tempout_path = Path(tempout_path)
    try:
        # Write content to the temporary file
        with os.fdopen(temp_fd, "w") as tmp:
            tmp.write("{key:value")
        cli(inline_args=[str(temp_path), "--indent", "0", "--ensure_ascii"])
        captured = capsys.readouterr()
        assert captured.out == '{\n"key": "value"\n}\n'

        # Test the output option
        cli(inline_args=[str(temp_path), "--indent", "0", "-o", str(tempout_path)])
        with tempout_path.open() as tmp:
            out = tmp.read()
        assert out == '{\n"key": "value"\n}'

        # Test the inline option
        cli(inline_args=[str(temp_path), "--indent", "0", "-i"])
        with temp_path.open() as tmp:
            out = tmp.read()
        assert out == '{\n"key": "value"\n}'

    finally:
        # Clean up - delete the temporary file
        temp_path.unlink()
        tempout_path.unlink()

    # Prepare a JSON string that needs to be repaired.
    test_input = "{key:value"
    # Expected output when running cli with --indent 0.
    expected_output = '{\n"key": "value"\n}\n'
    # Patch sys.stdin so that cli() reads from it instead of a file.
    with patch("sys.stdin", new=io.StringIO(test_input)):
        cli(inline_args=["--indent", "0"])
    captured = capsys.readouterr()
    assert captured.out == expected_output


def test_cli_inline_requires_filename(capsys):
    """cli() should exit with an error when --inline is passed without a filename."""
    with pytest.raises(SystemExit) as exc:
        cli(inline_args=["--inline"])
    captured = capsys.readouterr()
    assert captured.err.strip() == "Error: Inline mode requires a filename"
    assert exc.value.code != 0


def test_cli_inline_and_output_error(tmp_path, capsys):
    """cli() should exit with an error when --inline and --output are used together."""
    outfile = tmp_path / "out.json"
    with pytest.raises(SystemExit) as exc:
        cli(inline_args=["dummy.json", "--inline", "--output", str(outfile)])
    captured = capsys.readouterr()
    assert captured.err.strip() == "Error: You cannot pass both --inline and --output"
    assert exc.value.code != 0


def test_cli_schema_file_guides_repair(tmp_path, capsys):
    pytest.importorskip("jsonschema")
    schema_path = tmp_path / "schema.json"
    schema = {
        "type": "object",
        "properties": {"value": {"type": "integer"}},
        "required": ["value"],
    }
    schema_path.write_text(json.dumps(schema))
    input_path = tmp_path / "input.json"
    input_path.write_text('{"value": }')

    cli(inline_args=[str(input_path), "--indent", "0", "--schema", str(schema_path)])
    captured = capsys.readouterr()
    assert captured.out == '{\n"value": 0\n}\n'


def test_cli_schema_applies_to_valid_json(tmp_path, capsys):
    pytest.importorskip("jsonschema")
    schema_path = tmp_path / "schema.json"
    schema = {
        "type": "object",
        "properties": {"value": {"type": "integer"}},
        "required": ["value"],
    }
    schema_path.write_text(json.dumps(schema))
    input_path = tmp_path / "input.json"
    input_path.write_text('{"value": "1"}')

    cli(inline_args=[str(input_path), "--indent", "0", "--schema", str(schema_path)])
    captured = capsys.readouterr()
    assert captured.out == '{\n"value": 1\n}\n'

    cli(
        inline_args=[
            str(input_path),
            "--indent",
            "0",
            "--schema",
            str(schema_path),
            "--skip-json-loads",
        ]
    )
    captured = capsys.readouterr()
    assert captured.out == '{\n"value": 1\n}\n'


def test_cli_schema_model_guides_repair(tmp_path, capsys, monkeypatch):
    pytest.importorskip("jsonschema")
    pydantic = pytest.importorskip("pydantic")
    version = getattr(pydantic, "VERSION", getattr(pydantic, "__version__", "0"))
    if int(version.split(".")[0]) < 2:
        pytest.skip("pydantic v2 required")

    module_path = tmp_path / "schema_model.py"
    module_path.write_text("from pydantic import BaseModel\n\n\nclass SchemaModel(BaseModel):\n    value: int\n")
    monkeypatch.syspath_prepend(tmp_path)

    input_path = tmp_path / "input.json"
    input_path.write_text('{"value": "1"}')

    cli(
        inline_args=[
            str(input_path),
            "--indent",
            "0",
            "--schema-model",
            "schema_model:SchemaModel",
            "--skip-json-loads",
        ]
    )
    captured = capsys.readouterr()
    assert captured.out == '{\n"value": 1\n}\n'


def test_cli_schema_and_strict_error(tmp_path, capsys):
    schema_path = tmp_path / "schema.json"
    schema_path.write_text(json.dumps({"type": "integer"}))
    input_path = tmp_path / "input.json"
    input_path.write_text('{"value": }')

    with pytest.raises(SystemExit) as exc:
        cli(inline_args=[str(input_path), "--schema", str(schema_path), "--strict"])
    captured = capsys.readouterr()
    assert "schema" in captured.err.lower()
    assert exc.value.code != 0


def test_cli_schema_and_schema_model_are_mutually_exclusive(tmp_path, capsys):
    schema_path = tmp_path / "schema.json"
    schema_path.write_text(json.dumps({"type": "integer"}))

    with pytest.raises(SystemExit) as exc:
        cli(
            inline_args=[
                "dummy.json",
                "--schema",
                str(schema_path),
                "--schema-model",
                "schema_model:SchemaModel",
            ]
        )
    captured = capsys.readouterr()
    assert "schema" in captured.err.lower()
    assert exc.value.code != 0



import copy
from typing import Any, ClassVar

import pytest

from src.json_repair import repair_json
from src.json_repair.schema_repair import (
    SchemaRepairer,
    load_schema_model,
    normalize_missing_values,
    schema_from_input,
)
from src.json_repair.utils.constants import MISSING_VALUE


def test_missing_value_deepcopy():
    assert copy.deepcopy(MISSING_VALUE) is MISSING_VALUE


def test_normalize_missing_values_nested_and_invalid():
    assert normalize_missing_values(MISSING_VALUE) == ""
    assert normalize_missing_values({"a": MISSING_VALUE, "b": [MISSING_VALUE, 1]}) == {"a": "", "b": ["", 1]}
    with pytest.raises(ValueError, match="Object keys must be strings"):
        normalize_missing_values({1: "a"})
    with pytest.raises(ValueError, match="JSON compatible"):
        normalize_missing_values(object())


def test_load_schema_model_errors_and_success(tmp_path, monkeypatch):
    with pytest.raises(ValueError, match="Schema model must be in the form"):
        load_schema_model("invalid")

    module_path = tmp_path / "schema_mod.py"
    module_path.write_text("class SchemaModel:\n    pass\n")
    monkeypatch.syspath_prepend(tmp_path)

    assert load_schema_model("schema_mod:SchemaModel").__name__ == "SchemaModel"
    with pytest.raises(ValueError, match="not found"):
        load_schema_model("schema_mod:Missing")


def test_schema_from_input_basic_and_invalid():
    assert schema_from_input({"type": "string"}) == {"type": "string"}
    assert schema_from_input(True) is True
    assert schema_from_input(False) is False
    with pytest.raises(ValueError, match="Schema must be a JSON Schema"):
        schema_from_input(1)


def test_schema_from_input_model_defaults_and_required():
    pydantic = pytest.importorskip("pydantic")
    version = getattr(pydantic, "VERSION", getattr(pydantic, "__version__", "0"))
    if int(version.split(".")[0]) < 2:
        pytest.skip("pydantic v2 required")

    class DummyField:
        def __init__(self, default, default_factory=None, required=False):
            self.default = default
            self.default_factory = default_factory
            self._required = required

        def is_required(self):
            return self._required

    class DummyModel:
        @staticmethod
        def model_json_schema():
            return {"properties": {"name": "not-a-dict"}}

        model_fields: ClassVar[dict[str, DummyField]] = {
            "name": DummyField(default="x"),
            "items": DummyField(default=None, default_factory=lambda: ["a"]),
            "required_field": DummyField(default="y", required=True),
        }

    schema = schema_from_input(DummyModel)
    assert isinstance(schema, dict)
    assert schema["properties"]["name"]["default"] == "x"
    assert schema["properties"]["items"]["default"] == ["a"]
    assert "required_field" not in schema["properties"]

    class DummyModelWithDefaults:
        @staticmethod
        def model_json_schema():
            return {"properties": {"name": {"default": "keep"}}}

        model_fields: ClassVar[dict[str, DummyField]] = {"name": DummyField(default="x")}

    schema_with_default = schema_from_input(DummyModelWithDefaults)
    assert isinstance(schema_with_default, dict)
    assert schema_with_default["properties"]["name"]["default"] == "keep"


def test_schema_from_input_non_dict_properties():
    pydantic = pytest.importorskip("pydantic")
    version = getattr(pydantic, "VERSION", getattr(pydantic, "__version__", "0"))
    if int(version.split(".")[0]) < 2:
        pytest.skip("pydantic v2 required")

    class DummyField:
        def __init__(self, default):
            self.default = default
            self.default_factory = None

        def is_required(self):
            return False

    class DummyModel:
        @staticmethod
        def model_json_schema():
            return {"properties": []}

        model_fields: ClassVar[dict[str, DummyField]] = {"value": DummyField(default=1)}

    schema = schema_from_input(DummyModel)
    assert isinstance(schema, dict)
    assert schema["properties"]["value"]["default"] == 1


def test_schema_from_input_requires_pydantic_v2(monkeypatch):
    pydantic = pytest.importorskip("pydantic")
    monkeypatch.setattr(pydantic, "VERSION", "1.0")

    class DummyModel:
        @staticmethod
        def model_json_schema():
            return {}

    with pytest.raises(ValueError, match="pydantic v2"):
        schema_from_input(DummyModel)


def test_schema_repairer_validate_and_prepare():
    pytest.importorskip("jsonschema")
    repairer = SchemaRepairer({}, [])
    repairer.validate(1, True)
    with pytest.raises(ValueError, match="Schema does not allow"):
        repairer.validate(1, False)
    repairer.validate(1, {"type": "integer"})
    with pytest.raises(ValueError, match="is not of type"):
        repairer.validate("x", {"type": "integer"})

    schema = {
        "type": "array",
        "items": [{"type": "integer"}, {"type": "string"}],
        "additionalItems": False,
    }
    repairer.validate([1, "a"], schema)
    prepared = repairer._prepare_schema_for_validation(schema)
    assert prepared["prefixItems"]
    assert prepared["items"] is False

    schema2 = {
        "items": [{"type": "integer"}],
        "additionalItems": {"type": "string"},
    }
    prepared2 = repairer._prepare_schema_for_validation(schema2)
    assert prepared2["items"] == {"type": "string"}

    invalid_schema: Any = True
    with pytest.raises(ValueError, match="Schema must be an object"):
        repairer._prepare_schema_for_validation(invalid_schema)


def test_schema_repairer_resolve_schema_and_refs():
    root = {"defs": {"node": {"type": "string"}}, "flag": True, "flag_false": False, "bad": 1}
    repairer = SchemaRepairer(root, None)
    assert repairer.resolve_schema(None) is True
    assert repairer.resolve_schema(False) is False
    assert repairer.resolve_schema({"$ref": "#/defs/node"}) == {"type": "string"}
    assert repairer.resolve_schema({"$ref": "#/flag"}) is True
    assert repairer.resolve_schema({"$ref": "#/flag_false"}) is False
    invalid_schema: Any = "nope"
    with pytest.raises(ValueError, match="Schema must be an object"):
        repairer.resolve_schema(invalid_schema)
    with pytest.raises(ValueError, match="Schema keys must be strings"):
        repairer.resolve_schema({1: "bad"})

    with pytest.raises(ValueError, match="Unsupported \\$ref"):
        repairer._resolve_ref("http://example.com")
    with pytest.raises(ValueError, match="Unresolvable \\$ref"):
        repairer._resolve_ref("#/missing")
    with pytest.raises(ValueError, match="Unresolvable \\$ref"):
        repairer._resolve_ref("#/bad")


def test_schema_repairer_copy_json_value_edges():
    repairer = SchemaRepairer({}, [])
    assert repairer._copy_json_value({"a": {"b": 1}}, "$", "default") == {"a": {"b": 1}}
    with pytest.raises(ValueError, match="non-string key"):
        repairer._copy_json_value({1: "a"}, "$", "default")
    with pytest.raises(ValueError, match="not JSON compatible"):
        repairer._copy_json_value(object(), "$", "default")


def test_schema_repairer_object_and_array_helpers():
    repairer = SchemaRepairer({}, None)
    assert repairer.is_object_schema({"type": "object"}) is True
    assert repairer.is_object_schema({"type": ["null", "object"]}) is True
    assert repairer.is_object_schema({"properties": {}}) is True
    assert repairer.is_object_schema({"type": "string"}) is False

    assert repairer.is_array_schema({"type": "array"}) is True
    assert repairer.is_array_schema({"type": ["array", "null"]}) is True
    assert repairer.is_array_schema({"items": {"type": "string"}}) is True
    assert repairer.is_array_schema({"type": "object"}) is False

    assert repairer.is_object_schema(True) is False
    assert repairer.is_array_schema(False) is False


def test_repair_value_missing_and_unions():
    repairer = SchemaRepairer({}, [])
    assert repairer.repair_value(MISSING_VALUE, {"const": 1}, "$") == 1
    assert repairer.repair_value(MISSING_VALUE, {"enum": [2, 3]}, "$") == 2
    assert repairer.repair_value(MISSING_VALUE, {"default": "x"}, "$") == "x"
    assert repairer.repair_value(MISSING_VALUE, {"type": "string"}, "$") == ""

    schema_any = {"anyOf": [{"type": "integer"}, {"type": "string"}]}
    assert repairer.repair_value("1", schema_any, "$") == 1

    schema_one = {"oneOf": [{"type": "integer"}, {"type": "boolean"}]}
    with pytest.raises(ValueError, match="Expected boolean"):
        repairer.repair_value("nope", schema_one, "$")

    schema_all = {"allOf": [{"type": "string"}, {"enum": ["a"]}]}
    assert repairer.repair_value("a", schema_all, "$") == "a"
    with pytest.raises(ValueError, match="does not match enum"):
        repairer.repair_value("b", schema_all, "$")

    schema_union = {"type": ["integer", "string"]}
    assert repairer.repair_value("2", schema_union, "$") == 2

    with pytest.raises(ValueError, match="No schema matched"):
        repairer.repair_value("x", {"oneOf": []}, "$")
    with pytest.raises(ValueError, match="Expected boolean"):
        repairer.repair_value("x", {"type": ["integer", "boolean"]}, "$")
    with pytest.raises(ValueError, match="No schema type matched"):
        repairer.repair_value("x", {"type": []}, "$")

    schema_array_union = {"type": ["array", "string"], "items": {"type": "integer"}}
    assert repairer.repair_value(["1"], schema_array_union, "$") == [1]
    schema_obj_union = {"type": ["object", "string"], "properties": {"a": {"type": "integer"}}}
    assert repairer.repair_value({"a": "1"}, schema_obj_union, "$") == {"a": 1}


def test_repair_object_and_array_paths():
    repairer = SchemaRepairer({}, [])
    schema_obj = {
        "type": "object",
        "properties": {
            "a": {"type": "integer"},
            "b": {"type": "string", "default": "x"},
        },
        "required": ["a"],
        "patternProperties": {
            "^x": {"type": "integer"},
            "1$": {"type": "integer"},
        },
        "additionalProperties": False,
    }
    value = {"a": "1", "x1": "2", "extra": "drop"}
    assert repairer.repair_value(value, schema_obj, "$") == {"a": 1, "b": "x", "x1": 2}
    with pytest.raises(ValueError, match="Missing required properties"):
        repairer.repair_value({}, schema_obj, "$")

    schema_obj_extra = {"type": "object", "additionalProperties": {"type": "integer"}}
    assert repairer.repair_value({"a": "1"}, schema_obj_extra, "$") == {"a": 1}

    schema_min_props = {"type": "object", "minProperties": 1}
    with pytest.raises(ValueError, match="minProperties"):
        repairer.repair_value({}, schema_min_props, "$")

    schema_bad_props = {"type": "object", "properties": [], "patternProperties": []}
    assert repairer.repair_value({}, schema_bad_props, "$") == {}
    with pytest.raises(ValueError, match="Expected object"):
        repairer.repair_value([], {"type": "object"}, "$")

    schema_array = {
        "type": "array",
        "items": [{"type": "integer"}],
        "additionalItems": False,
    }
    assert repairer.repair_value([1, 2], schema_array, "$") == [1]
    schema_tuple = {"type": "array", "items": [{"type": "integer"}, {"type": "string"}]}
    assert repairer.repair_value([1], schema_tuple, "$") == [1]

    schema_array_extra = {
        "type": "array",
        "items": [{"type": "integer"}],
        "additionalItems": {"type": "string"},
    }
    assert repairer.repair_value([1, 2], schema_array_extra, "$") == [1, "2"]

    schema_array_open = {
        "type": "array",
        "items": [{"type": "integer"}],
        "additionalItems": True,
    }
    assert repairer.repair_value([1, "x"], schema_array_open, "$") == [1, "x"]

    schema_array_items = {"type": "array", "items": {"type": "integer"}}
    assert repairer.repair_value(["1", 2], schema_array_items, "$") == [1, 2]

    schema_array_wrap = {"type": "array"}
    assert repairer.repair_value("a", schema_array_wrap, "$") == ["a"]

    schema_min_items = {"type": "array", "minItems": 1}
    with pytest.raises(ValueError, match="minItems"):
        repairer.repair_value([], schema_min_items, "$")


def test_fill_missing_and_coerce_scalar_paths():
    repairer = SchemaRepairer({}, [])
    assert repairer._fill_missing({"type": "integer"}, "$") == 0
    assert repairer._fill_missing({"type": "number"}, "$") == 0
    assert repairer._fill_missing({"type": "boolean"}, "$") is False
    assert repairer._fill_missing({"type": "null"}, "$") is None
    assert repairer._fill_missing({"type": "array"}, "$") == []
    assert repairer._fill_missing({"type": "object"}, "$") == {}
    assert repairer._fill_missing({"type": ["string", "integer"]}, "$") == ""
    assert repairer._fill_missing({"properties": {"a": {"type": "string"}}}, "$") == {}
    assert repairer._fill_missing({"items": {"type": "string"}}, "$") == []
    with pytest.raises(ValueError, match="requires at least"):
        repairer._fill_missing({"type": "array", "minItems": 1}, "$")
    with pytest.raises(ValueError, match="requires at least"):
        repairer._fill_missing({"type": "object", "minProperties": 1}, "$")
    with pytest.raises(ValueError, match="Cannot infer missing value"):
        repairer._fill_missing({"type": "custom"}, "$")
    with pytest.raises(ValueError, match="has no values"):
        repairer._fill_missing({"enum": []}, "$")
    with pytest.raises(ValueError, match="Cannot infer missing value"):
        repairer._fill_missing({"type": ["unsupported"]}, "$")

    assert repairer._coerce_scalar(1, "string", "$") == "1"
    assert repairer._coerce_scalar("2", "integer", "$") == 2
    assert repairer._coerce_scalar("2.0", "integer", "$") == 2
    assert repairer._coerce_scalar(2.0, "integer", "$") == 2
    assert repairer._coerce_scalar("2.5", "number", "$") == 2.5
    assert repairer._coerce_scalar("true", "boolean", "$") is True
    assert repairer._coerce_scalar("false", "boolean", "$") is False
    assert repairer._coerce_scalar(None, "null", "$") is None

    with pytest.raises(ValueError, match="Expected string"):
        repairer._coerce_scalar(True, "string", "$")
    with pytest.raises(ValueError, match="Expected integer"):
        repairer._coerce_scalar(True, "integer", "$")
    with pytest.raises(ValueError, match="Expected integer"):
        repairer._coerce_scalar(2.5, "integer", "$")
    with pytest.raises(ValueError, match="Expected integer"):
        repairer._coerce_scalar("2.5", "integer", "$")
    with pytest.raises(ValueError, match="Expected integer"):
        repairer._coerce_scalar({}, "integer", "$")
    with pytest.raises(ValueError, match="Expected number"):
        repairer._coerce_scalar(True, "number", "$")
    with pytest.raises(ValueError, match="Expected number"):
        repairer._coerce_scalar("nope", "number", "$")
    with pytest.raises(ValueError, match="Expected number"):
        repairer._coerce_scalar([], "number", "$")
    with pytest.raises(ValueError, match="Expected boolean"):
        repairer._coerce_scalar("yes", "boolean", "$")
    with pytest.raises(ValueError, match="Expected null"):
        repairer._coerce_scalar("x", "null", "$")
    with pytest.raises(ValueError, match="Unsupported schema type"):
        repairer._coerce_scalar("x", "unsupported", "$")

    assert repairer.repair_value({"a": MISSING_VALUE}, {}, "$") == {"a": ""}
    assert repairer.repair_value({"a": MISSING_VALUE}, {"allOf": []}, "$") == {"a": ""}
    assert repairer.repair_value({"a": "1"}, {"properties": {"a": {"type": "integer"}}}, "$") == {"a": 1}
    assert repairer.repair_value(["1"], {"items": {"type": "integer"}}, "$") == [1]
    with pytest.raises(ValueError, match="Schema does not allow"):
        repairer.repair_value(1, False, "$")


def test_apply_enum_const_mismatch_raises():
    repairer = SchemaRepairer({}, [])
    with pytest.raises(ValueError, match="does not match const"):
        repairer.repair_value("b", {"const": "a"}, "$")
    with pytest.raises(ValueError, match="does not match enum"):
        repairer.repair_value("b", {"enum": ["a"]}, "$")


def test_repair_json_valid_empty_string_returns_empty():
    assert repair_json('""') == ""



import pytest

from src.json_repair import repair_json
from src.json_repair.json_parser import JSONParser
from src.json_repair.schema_repair import SchemaRepairer
from src.json_repair.utils.json_context import ContextValues


def parse_object_direct(raw, schema, *, strict=False, context=None):
    parser = JSONParser(raw, None, False, 0, False, strict)
    repairer = SchemaRepairer(schema if isinstance(schema, dict) else {}, None)
    parser.schema_repairer = repairer
    if context is not None:
        parser.context.set(context)
    parser.index = 1
    return parser.parse_object(schema, "$")


def parse_array_direct(raw, schema):
    parser = JSONParser(raw, None, False, 0, False, False)
    repairer = SchemaRepairer(schema if isinstance(schema, dict) else {}, None)
    parser.schema_repairer = repairer
    parser.index = 1
    return parser.parse_array(schema, "$")


def test_parse_object_schema_true_false_and_non_object():
    assert parse_object_direct("{}", True) == {}
    with pytest.raises(ValueError, match="Schema does not allow"):
        parse_object_direct("{}", False)
    assert parse_object_direct("{}", {"type": "string"}) == {}


def test_parse_object_schema_property_fallbacks_and_stray_colon():
    schema = {
        "type": "object",
        "properties": [],
        "patternProperties": [],
        "additionalProperties": True,
    }
    assert parse_object_direct("{:a:1}", schema) == {"a": 1}


def test_parse_object_schema_invalid_property_schema_raises():
    schema = {"type": "object", "properties": {"a": "nope"}}
    with pytest.raises(ValueError, match="Schema must be an object"):
        parse_object_direct('{"a": 1}', schema)


def test_parse_object_schema_invalid_pattern_schema_raises():
    schema = {"type": "object", "patternProperties": {"^a": "nope"}}
    with pytest.raises(ValueError, match="Schema must be an object"):
        parse_object_direct('{"a": 1}', schema)


def test_parse_object_schema_invalid_pattern_extra_schema_raises():
    schema = {
        "type": "object",
        "patternProperties": {"^a": {"type": "integer"}, "a$": "nope"},
    }
    with pytest.raises(ValueError, match="Schema must be an object"):
        parse_object_direct('{"a": 1}', schema)


def test_parse_array_schema_missing_object_brace():
    schema = {
        "type": "array",
        "items": {
            "type": "object",
            "properties": {"a": {"type": "integer"}},
            "required": ["a"],
        },
    }
    assert repair_json('["a": 1]', schema=schema, skip_json_loads=True, return_objects=True) == [{"a": 1}]


def test_parse_object_schema_pattern_extra_schemas():
    schema = {
        "type": "object",
        "patternProperties": {
            "^x": {"type": "integer"},
            "1$": {"type": "integer"},
        },
    }
    assert repair_json('{"x1": "2"}', schema=schema, skip_json_loads=True, return_objects=True) == {"x1": 2}


def test_parse_object_schema_drop_property_and_additional_schema():
    schema_drop = {
        "type": "object",
        "properties": {"a": {"type": "integer"}},
        "additionalProperties": False,
    }
    assert repair_json('{"a": 1, "extra": "drop"}', schema=schema_drop, skip_json_loads=True, return_objects=True) == {
        "a": 1
    }

    schema_extra = {
        "type": "object",
        "additionalProperties": {"type": "integer"},
    }
    assert repair_json('{"a": "2"}', schema=schema_extra, skip_json_loads=True, return_objects=True) == {"a": 2}


def test_parse_object_schema_closing_array_bracket_and_extra_brace():
    schema = {
        "type": "array",
        "items": {
            "type": "object",
            "properties": {"a": {"type": "integer"}},
            "required": ["a"],
        },
    }
    assert repair_json('[{"a": 1]', schema=schema, skip_json_loads=True, return_objects=True) == [{"a": 1}]

    schema_obj = {"type": "object", "additionalProperties": True}
    parser = JSONParser('{"a": 1}}', None, False, 0, False, False)
    repairer = SchemaRepairer(schema_obj, None)
    parser.schema_repairer = repairer
    parser.context.set(ContextValues.ARRAY)
    parser.index = 1
    assert parser.parse_object(schema_obj, "$") == {"a": 1}


def test_parse_object_schema_empty_object_falls_back_to_array():
    schema = {"type": "object", "additionalProperties": True}
    assert parse_object_direct("{,,}", schema) == []


def test_parse_array_schema_true_false_and_non_array():
    assert parse_array_direct("[1]", True) == [1]
    with pytest.raises(ValueError, match="Schema does not allow"):
        parse_array_direct("[1]", False)
    assert parse_array_direct("[1]", {"type": "object"}) == [1]


def test_parse_array_schema_items_and_additional_items():
    schema_drop = {"type": "array", "items": [{"type": "integer"}], "additionalItems": False}
    assert parse_array_direct("[1, 2]", schema_drop) == [1]

    schema_extra = {
        "type": "array",
        "items": [{"type": "integer"}],
        "additionalItems": {"type": "integer"},
    }
    assert parse_array_direct('[1, "2"]', schema_extra) == [1, 2]

    schema_open = {
        "type": "array",
        "items": [{"type": "integer"}],
        "additionalItems": True,
    }
    assert parse_array_direct('[1, "x"]', schema_open) == [1, "x"]

    schema_items = {"type": "array", "items": {"type": "integer"}}
    assert parse_array_direct('["1"]', schema_items) == [1]

    schema_any = {"type": "array"}
    assert parse_array_direct("[true]", schema_any) == [True]


def test_parse_array_schema_invalid_items_schema_raises():
    schema = {"type": "array", "items": ["nope"]}
    with pytest.raises(ValueError, match="Schema must be an object"):
        parse_array_direct("[1]", schema)


def test_parse_json_with_schema_branches():
    schema = {"type": "array", "items": {"type": "integer"}}
    parser = JSONParser("[1]", None, False, 0, False, False)
    repairer = SchemaRepairer(schema, None)
    parser.schema_repairer = repairer
    assert parser.parse_json(schema, "$") == [1]

    parser = JSONParser('"1"', None, False, 0, False, False)
    parser.schema_repairer = repairer
    parser.context.set(ContextValues.ARRAY)
    assert parser.parse_json({"type": "integer"}, "$") == 1

    parser = JSONParser("1", None, False, 0, False, False)
    parser.schema_repairer = repairer
    parser.context.set(ContextValues.ARRAY)
    assert parser.parse_json({"type": "integer"}, "$") == 1

    parser = JSONParser("# comment", None, False, 0, False, False)
    parser.schema_repairer = repairer
    assert parser.parse_json({"type": "string"}, "$") == ""

    parser = JSONParser("", None, False, 0, False, False)
    parser.schema_repairer = repairer
    assert parser.parse_json({"type": "string"}, "$") == ""

    parser = JSONParser('"x"', None, False, 0, False, False)
    parser.schema_repairer = repairer
    parser.context.set(ContextValues.ARRAY)
    assert parser.parse_json(True, "$") == "x"
    with pytest.raises(ValueError, match="Schema does not allow"):
        parser.parse_json(False, "$")

    parser = JSONParser('@{"a": 1}', None, False, 0, False, False)
    parser.schema_repairer = SchemaRepairer({"type": "object"}, None)
    assert parser.parse_json({"type": "object"}, "$") == {"a": 1}


def test_schema_ref_to_true_short_circuits():
    schema = {"flag": True}
    repairer = SchemaRepairer(schema, None)

    parser = JSONParser("[1]", None, False, 0, False, False)
    parser.schema_repairer = repairer
    parser.index = 1
    assert parser.parse_array({"$ref": "#/flag"}, "$") == [1]

    parser = JSONParser('{"a": 1}', None, False, 0, False, False)
    parser.schema_repairer = repairer
    parser.index = 1
    assert parser.parse_object({"$ref": "#/flag"}, "$") == {"a": 1}

    parser = JSONParser("1", None, False, 0, False, False)
    parser.schema_repairer = repairer
    parser.context.set(ContextValues.ARRAY)
    assert parser.parse_json({"$ref": "#/flag"}, "$") == 1
